#  simple text parser classes for text comparing function
=begin
    Copyright (C) 2008  Tomáš Gregorovič, _tom_ at centrum.cz

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
=end

require 'strscan'
require 'textsimilarity'

class ParsedPatternQuery
  def initialize(words)
    @words = words
	@wordpos = {}
	@foundcount = 0
	@from = 0
	@requestedcount = 0
    @returnedcount = 1
    @type = 'all'
	
    @words.each_index { |i| w = @words[i]
      a = @wordpos[w.hash]
      if a != nil then
        a.push(i)
      else
        @wordpos[w.hash] = [i]
      end
	}
  end
  def words # array of string
    @words
  end
  def wordpos(word) # hash of array of pattern pos
    @wordpos[word.hash]
  end
  def foundcount
    @foundcount
  end
  def setfoundcount(count)
    @foundcount = count
  end
  def requestedcount
    @requestedcount
  end
  def setrequestedcount(count)
    @requestedcount = count
  end
  def returnedcount
    @returnedcount
  end
  def setreturnedcount(count)
    @returnedcount = count
    if count == 0 and @type == 'all' then
      @type = 'any'
      @returnedcount = 1
    end
  end
  def from
    @from
  end
  def setfrom(from)
    @from = from
  end
  def type
    @type
  end
  def exhausted?
    return (@returnedcount == 0 and @type == 'any')
  end
end

class ParsedPattern
  def initialize(text, maxchunklen = 40)
    @text = text
    @words = []
    @wordpos = {}
	@queries = [] 
    
    scanner = StringScanner.new(text)
    index = 0
    while not scanner.eos?
      w = scanner.scan(/\S+/)
      
      if w != nil then
        @words.push(w)
        a = @wordpos[w.hash]
        if a != nil then
          a.push(index)
        else
          @wordpos[w.hash] = [index]
        end
        index = index + 1
      end
      
      scanner.skip(/\s+/) 
    end
    
	# create chunks of max 800 bytes for search query with respect to sentences
	s = 0
    len = 0
    i = 0
    while i < @words.length 
      w = @words[i]
	  wordlen = w.length * 3 + 3
      if (len + wordlen > 800) or (i - s > maxchunklen - 1) then
        @queries.push(ParsedPatternQuery.new(@words.slice(s, i - s)))
        len = 0
		s = [i - 8, s + 1].max
        i = s
	  else
        len = len + wordlen
        i = i + 1
	  end
	end
    
    if @words.length - s >= 8 then
	  @queries.push(ParsedPatternQuery.new(@words.slice(s, @words.length - s)))
    end
  end
  def text
    @text
  end
  def words # array of string
    @words
  end
  def queries # array of ParsedPatternQueries
    @queries
  end
  def wordpos(word) # hash of array of pattern pos
    @wordpos[word.hash]
  end
end

class ParsedDocument
  def initialize(pattern, words)
    @text = text
    @words = words
    allwordmatches = []
    @wordsimilarityindex = -1
    
    chunklength = 7
    minchunkwordcount = 5
	
    @words.each { |w|      
      allwordmatches.push(pattern.wordpos(w))
    }
	
	# get siomilarity index for each word
	#  this time expensive function is called as c extension
	
	extension = TextSimilarityExtension.new(allwordmatches)
	matches = extension.result
	  
    # check count of similar words in each chunk
    @wordmatches = Array.new(@words.length, 0.0)
    chunkwordcount = 0
    checkwordlist = []
    @words.each_index { |i|
      if matches[i] > 0 then
        chunkwordcount = chunkwordcount + 1
      end
      if i > 8 and matches[i - chunklength] > 0 then
        chunkwordcount = chunkwordcount - 1
        checkwordlist.delete(i - chunklength)
      end
      
      if chunkwordcount >= minchunkwordcount then
        @wordmatches[i] = matches[i]
        checkwordlist.each { |p| @wordmatches[p] = matches[p]}
        checkwordlist.clear
      else
        checkwordlist.push(i)
      end
    }

    # count the index of similarity
    @similarityindex = 0.0
    @similarwordscount = 0
    @words.each_index { |i|      
      if @wordmatches[i] > 0 then
        @similarityindex = @similarityindex + @wordmatches[i]
        @similarwordscount = @similarwordscount + 1
      end
    }
  end
  def text
    @text
  end
  def words # array of string
    @words
  end
  def wordmatches # array of word pattern similarity
    @wordmatches
  end
  def similarityindex
    @similarityindex
  end
  def similarityratio
    if @similarwordscount > 0 then
      @similarityindex / @similarwordscount
    else
      0
    end
  end
  def similarwordscount
    @similarwordscount
  end
  def sample(maxchars = 200)
    result = ''
    addellipsis = false
    
    words.each_index { |i|
      if @wordmatches[i] > 0 then
        if addellipsis then
          result = result + ' …'
          maxchars = maxchars - 2
        end
        result = result + ' ' + @words[i]
        maxchars = maxchars - (@words[i].length + 1)
        addellipsis = false
      else
        addellipsis = true
      end
      
      if maxchars < 0 then break end
    }
    
    return result + ' …'
  end
end

class ParsedSummary<ParsedDocument
  def initialize(pattern, summary)
    wordlist = []
    if summary == nil then summary = '' end
    
    scanner = StringScanner.new(summary)
    while not scanner.eos?
      w = scanner.scan(/\S+/)
      
      if w != nil then
        if w != '...' then
          wordlist.push(w)
        else
          4.times { wordlist.push(' ') }
        end
      end
      
      scanner.skip(/\s+/) 
    end
    
    super(pattern, wordlist)
  end
end